/*
 * Copyright (C) 2008-2009 Alex Lasnier <alex@fepg.org>
 *
 * This file is part of NAScan
 *
 * NAScan is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * NAScan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NAScan.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ctype.h>

#include <libsi/section.h>
#include <libsi/descriptor.h>

#include "nitfilter.h"
#include "tools.h"


#define NIT_TID  0x40
#define NIT_PID  0x0010  // NA providers always use Nit pid 16 ?


//////////////////////////////////////////////////////////////////////////////


NitFilter::NitFilter(cList<TransponderSorter>* tl)
{
  transponderList = tl;
  gotNit = false;
  gotData = false;
  sections = 0;
  numSections = 0;
  
  Set(NIT_PID, NIT_TID);
}


//----------------------------------------------------------------------------

NitFilter::~NitFilter()
{

}


//----------------------------------------------------------------------------

void NitFilter::SetStatus(bool On)
{ 
  if (On)
  {
    gotNit = false;
    gotData = false;
    sections = 0;
    numSections = 0;
  }
  
  cFilter::SetStatus(On);  
}


//----------------------------------------------------------------------------

void NitFilter::Process(u_short pid, u_char tid, const u_char* data, int length)
{
  SI::NIT nit(data, false);
  if (!nit.CheckCRCAndParse())
    return;
  
  if ( GotSection(nit.getSectionNumber()) )
    return;
  
  gotData = true;
     
  dprint(L_NIT, "NIT: Received NIT (nid %d) %d of %d.", nit.getNetworkId(), nit.getSectionNumber()+1, nit.getLastSectionNumber()+1);
  
  if (!numSections)
    numSections = nit.getLastSectionNumber() + 1;
  
  SeenSection(nit.getSectionNumber());
  
  // Get network name
  if (nit.getSectionNumber() == 0)
  {
    SI::Descriptor *d;
    for (SI::Loop::Iterator it; (d = nit.commonDescriptors.getNext(it)); ) {
      switch (d->getDescriptorTag()) {
        case SI::NetworkNameDescriptorTag: {
          SI::NetworkNameDescriptor* nnd = (SI::NetworkNameDescriptor*) d;
          char netBuf[nnd->name.getLength()+1];
          nnd->name.getText(netBuf, sizeof(netBuf));
          providerNames[nit.getNetworkId()] = netBuf;
          dprint(L_NIT, "NIT: Found network name %s for nid %d", netBuf, nit.getNetworkId());
        }
        break;
        default: ;
      }
      delete d;
    }
  }
  
  // Find transponders
  SI::NIT::TransportStream ts;
  for (SI::Loop::Iterator it; nit.transportStreamLoop.getNext(ts, it); ) 
  {
    SI::Descriptor *d = NULL;
    for (SI::Loop::Iterator it2; (d = ts.transportStreamDescriptors.getNext(it2)); ) 
    {
      switch (d->getDescriptorTag()) 
      {
        case SI::ServiceListDescriptorTag:
        /*{
          SI::ServiceListDescriptor* sld = (SI::ServiceListDescriptor*) d;
          SI::ServiceListDescriptor::Service ser;
          for (SI::Loop::Iterator it3; sld->serviceLoop.getNext(ser, it3); ) {
            dprint(L_NIT, "NIT: SLD sid %d, type 0x%02X", ser.getServiceId(), ser.getServiceType());
          }
        }*/
        break;
        
        case SI::SatelliteDeliverySystemDescriptorTag: 
        {
          SI::SatelliteDeliverySystemDescriptor* sd = (SI::SatelliteDeliverySystemDescriptor*) d;

          int Frequency = BCD2INT(sd->getFrequency()) / 100;
          char Polarization = sd->getPolarization() == 2 ? 'H' : 'V'; //XXX: Determine if we should use H/V or R/L
          
          //dprint(L_NIT, "NIT: %d MHz", Frequency);
          
          // Sometimes sd->getWestEastFlag() is wrong ?!
          int Source = cSource::FromData(cSource::stSat, BCD2INT(sd->getOrbitalPosition()) );//, sd->getWestEastFlag());
          int SymbolRate = BCD2INT(sd->getSymbolRate()) / 10; 
          
          //dprint(L_NIT | L_MSG, "NIT: Found transponder (%s %d:%c)", *cSource::ToString(Source), Frequency, toupper(Polarization));
          
          // Values only valid for VDR 1.7.0 ?
          //static int CodeRates[]   = { 0, 12, 23, 34, 56, 78, 89, 35, 45, 910, 999, 999, 999, 999, 999, 0 };
            
          // int CodeRate   = MapToDriver(CodeRates[sd->getFecInner()], CoderateValues);
          int CodeRate   = MapToDriver(999, CoderateValues); // Auto
          
#if VDRVERSNUM < 10700

          int System   = 0;
          int RollOff  = 0;
          int Modulation = 0; // not used
          // TODO: Use modulation when genpix patch is applied. Can we detect the patch?

#else     
      
          int System   = MapToDriver(sd->getModulationSystem() ? 1 : 0, SystemValues);
          
          static int RollOffs[] = { 35, 25, 20, 0 };
          int RollOff  = MapToDriver(sd->getModulationSystem() ? RollOffs[sd->getRollOff()] : 0, RollOffValues);
          
          static int Modulations[] = {      998,        2,        5,        16 };
                                  // { MOD_AUTO, MOD_QPSK, MOD_8PSK, MOD_QAM16 };
          int Modulation = MapToDriver(Modulations[sd->getModulationType()], ModulationValues);        
#endif      
      
          //dprint(L_NIT, "NIT: cr %d, mod %d, sys %d, ro %d", CodeRate, Modulation, System, RollOff);
          
          // For spot beams, the TSID we get here is useless, but we will get the right one from the PAT.
          AddTransponder(Source, ts.getOriginalNetworkId(), ts.getTransportStreamId(), Frequency, Polarization, SymbolRate, CodeRate, Modulation, System, RollOff);                  
        }
        break;

        default:
          dprint(L_NIT, "NIT: Unsupported Delivery System Descriptor 0x%02X", d->getDescriptorTag()); 
        break;
      }
      
      delete d;
    }
  }
  
  
  if (GotAllSections()) {
    dprint(L_NIT, "NIT: Got all sections");
    
    gotNit = true;
    SetStatus(false);
  }
}


//----------------------------------------------------------------------------

bool NitFilter::GotSection(int i) const
{
  if (i < 0 || i >= numSections)
    return false;

  return (sections >> i) & 1;
}


//----------------------------------------------------------------------------  

bool NitFilter::GotAllSections(void) const
{
  uint64_t i = 1;
  i = i << numSections;

  return (sections == i-1);
}


//----------------------------------------------------------------------------  

void NitFilter::SeenSection(int i)
{
  if (i < 0 || i >= numSections)
    return;
    
  sections |= (1 << i);
}


//---------------------------------------------------------------------------- 

void NitFilter::AddTransponder(int Source, int Nid, int Tid, int Frequency, char Polarization, int Srate, int CoderateH, int Modulation, int System, int RollOff)
{
  
  for(TransponderSorter* ts = transponderList->First(); ts; ts = transponderList->Next(ts))
#if VDRVERSNUM < 10700
    if (ts->channel->Frequency() ==  Frequency)
#else 
    if (ts->channel->Frequency() ==  Frequency && ts->channel->Modulation() == Modulation)
#endif    
      return;
  
  // dprint(L_MSG, "NIT: Adding TS: F: %d   N: %d   T: %d   M: %d", Frequency, Nid, Tid, Modulation);
  
  cChannel* curTransponder = new cChannel();
  
#if VDRVERSNUM < 10700
  curTransponder->SetSatTransponderData(Source, Frequency, Polarization, Srate, CoderateH);
#else
  curTransponder->SetSatTransponderData(Source, Frequency, Polarization, Srate, CoderateH, Modulation, System, RollOff);
#endif
  
  curTransponder->SetId(Nid, Tid, 0, 0);
  
  transponderList->Add(new TransponderSorter(curTransponder));
}


//---------------------------------------------------------------------------- 

const char* NitFilter::GetNetworkName(uint16_t nid) const
{
  std::map<uint16_t, std::string>::const_iterator itr = providerNames.find(nid);
  
  return (itr != providerNames.end()) ? (*itr).second.c_str() : "";
}


//////////////////////////////////////////////////////////////////////////////


