/*
 * Copyright (C) 2008-2009 Alex Lasnier <alex@fepg.org>
 *
 * This file is part of NAScan
 *
 * NAScan is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * NAScan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NAScan.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include <stdarg.h>

#include <vdr/plugin.h>

#include "osd.h"
#include "tools.h"
#include "config.h"


#define SLEEP_TIME  100

#define SOURCES_FILE "sources.conf"


//////////////////////////////////////////////////////////////////////////////


ScanOsd::ScanOsd(void) : cOsdMenu("NA Channel Scanner"), cThread("NA Scan"), nitFilter(&transponderList), pmtFilter(&channelList)
{
  SetHelp("Cancel");
	
	dir = cPlugin::ConfigDirectory("nascan");

	numSources  = 0;
	needsUpdate = 0;
	
  Display();
  Start(); 
}


//----------------------------------------------------------------------------

ScanOsd::~ScanOsd(void) 
{
  Cancel(SLEEP_TIME+500);
}


//----------------------------------------------------------------------------

void ScanOsd::Action(void)
{
  if (!Load(dir, SOURCES_FILE))
    return;
  
  time_t startTime = time(NULL);
  ChannelSorter* lastChannel = NULL;
  
  // Start scanning the sources
  for (int i=0; i<numSources && Running(); i++)
  {
    char buf[74];
    snprintf(buf, 74, "--- Scanning %s %s -------------------------------------------------------------",
                       *cSource::ToString(sources[i]), Sources.Get(sources[i])->Description());
    AddLine(buf); 
            
    transponderList.Clear();
    
    // Tune initial transponder: get SDTs and NITs 
    cChannel* initial = GetInitial(i);

    cDevice* device = NULL;
    if (config.adapter == 0) // Auto
    {
#if VDRVERSNUM < 10500  
      device = cDevice::GetDevice(initial);
#else
      device = cDevice::GetDevice(initial, -1, false);
#endif  
    }
    else device = cDevice::GetDevice(config.adapter - 1);
    
    if (!device) {
      AddLine("ERROR: No device found.");
      return;
    }
    
    if (!Tune(device, initial, config.initialTuningTimeout)) {
      AddLine("ERROR: Unable to tune initial transponder.");
      continue;
    }
    
    delete initial; initial = NULL;
    
    AddLine("Downloading all SDTs/NITs, this may take a while...");
    
    sdtFilter.Attach(device);
    nitFilter.Attach(device);
    
    cCondWait::SleepMs(1000);
    
    int timeout = 0;
    while (Running() && !nitFilter.GotNit() && timeout <= config.nitTimeout) {
      cCondWait::SleepMs(SLEEP_TIME);
      if (config.nitTimeout && !nitFilter.GotData()) 
        timeout += SLEEP_TIME;
    }
    if (Running() && nitFilter.GotNit())
      AddLine("Received all NITs: found %d transponders", transponderList.Count());
    
    if (nitFilter.GotNit())
      while (Running() && !sdtFilter.GotSdts() && timeout <= config.sdtTimeout) {
        cCondWait::SleepMs(SLEEP_TIME);
        if (config.sdtTimeout && !sdtFilter.GotData()) 
          timeout += SLEEP_TIME;
      }
    sdtFilter.Detach();
    nitFilter.Detach();
    
    if (!Running()) break;
    
    if (!sdtFilter.GotSdts() || !nitFilter.GotNit()) {
      AddLine("ERROR: SDT/NIT Timeout (%d %d).", sdtFilter.GotSdts(), nitFilter.GotNit());
      AddLine("");
      continue;    
    }
    
    AddLine("Received all SDTs");

    
    // Scan each transponder
    transponderList.Sort();  

    for (TransponderSorter* ts = transponderList.First() ; ts && Running(); ts = transponderList.Next(ts)) 
    {
      pmtFilter.SetTransponder(ts->channel);
      
      if (!Tune(device, ts->channel, config.tuningTimeout))
        continue;
      
      pmtFilter.Attach(device);
      cCondWait::SleepMs(1000);
      
      // Wait for PMTs  
      timeout = 0;
      while (Running() && !pmtFilter.GotPMT() && timeout <= config.pmtTimeout) {
        cCondWait::SleepMs(SLEEP_TIME);
        if (config.pmtTimeout)
          timeout += SLEEP_TIME;
      }      
      if (!Running()) break;
      
      if (!pmtFilter.GotPMT()) {
        AddLine("ERROR: PMT Timeout. (%d)", pmtFilter.GotPAT());
        continue;
      }
      
      AddLine("  Found %d channels", pmtFilter.GotPMT());
      pmtFilter.Detach();
      
      cCondWait::SleepMs(100);
    }
    if (!Running()) break;

    int numChannels = 0;
    if (!lastChannel)
      lastChannel = channelList.First();
       
    for (ChannelSorter* cs = lastChannel; cs; )
    {
      cChannel* c = cs->channel;
        
      const char* name = names.Name(c->Nid(), c->Tid(), c->Sid());
      if (!name)
        name = sdtFilter.GetName(c->Nid(), c->Tid(), c->Sid());
        
      const char* shortName = sdtFilter.GetShortName(c->Nid(), c->Tid(), c->Sid());
      const char* networkName = config.useNetworkName ? nitFilter.GetNetworkName(c->Nid()) : "";
        
      if (name && shortName)
      {
        c->SetName(name, shortName, networkName);
        numChannels++;
        cs = channelList.Next(cs);
      }
      else 
      {
        ChannelSorter* prev = cs;
        cs = channelList.Next(cs);
        channelList.Del(prev);
      } 
    }
    lastChannel = channelList.Last();
    
          
    if (config.outputType == OT_SOURCES) // Different file for every source
    {
      FILE* outFile = NULL; 
      char* fn = NULL;
      asprintf(&fn, "%s/channels-%s.conf", dir, *cSource::ToString(sources[i]));
  
      outFile = fopen(fn, "w");
      free(fn); fn = NULL;
      if (outFile) {
        for (ChannelSorter* cs = channelList.First(); cs; cs = channelList.Next(cs))
        {
          cString chanText = cs->channel->ToText();
          fprintf(outFile, "%s", *chanText);
        }
        fclose(outFile);
      }
      
      channelList.Clear();
      lastChannel = NULL;
    }

    
    // TEMP --- START ---
/*
    asprintf(&fn, "%s/names-%s.conf", dir, *cSource::ToString(sources[i]));
    outFile = fopen(fn, "w");
    free(fn); fn = NULL;
    if (outFile) {
      for (ChannelSorter* cs = channelList.First(); cs; cs = channelList.Next(cs)) {
        cChannel* c = cs->channel;
        const char* name = sdtFilter.GetName(c->Nid(), c->Tid(), c->Sid());
        if (name) 
        {
          fprintf(outFile, "%s:%d:%d:%d\n", name , c->Sid(), c->Nid(), c->Tid());
        }
      }
      fclose(outFile);
    }  
*/      
    // TEMP --- END ---
    
    //transponderList.Clear();
    
    AddLine("Found a total of %d channels on %s", numChannels, Sources.Get(sources[i])->Description());
    AddLine("");
    
    cCondWait::SleepMs(100);
  }
  
  
  if (config.outputType == OT_COMPLETE) // Generate complete channels.conf
  {
    channelList.Sort();
    
    FILE* outFile = NULL; 
    char* fn = NULL;
    asprintf(&fn, "%s/channels.conf", dir);
  
    outFile = fopen(fn, "w");
    free(fn); fn = NULL;
    if (outFile) {
      for (ChannelSorter* cs = channelList.First(); cs; cs = channelList.Next(cs))
      {
        cChannel* c = cs->channel;
        cString chanText = c->ToText();
        fprintf(outFile, ":@%d\n%s", c->Number(), *chanText);
      }
      fclose(outFile);
    }
      
    channelList.Clear();
  }
  

  pmtFilter.Detach();
  if (Running())
    AddLine("Done (took %d seconds)", time(NULL) - startTime);
}


//----------------------------------------------------------------------------

eOSState ScanOsd::ProcessKey(eKeys Key)
{
  eOSState state = cOsdMenu::ProcessKey(Key);

  if (state == osUnknown)
  {
    state = osContinue;
    
    switch (Key)
    {
      case kBack:
        state = osBack;
      break;
      
      case kRed:
        if (Running()) {
          Cancel(SLEEP_TIME+500);
          AddLine("Scan cancelled");
        }
        else
          state = osBack;
      break;
      
      case kNone:
        if (needsUpdate) {
          Display();
          needsUpdate = false;
        }
      break;
      
      default:
      break;    
    }
  }  
  return state;  
}


//----------------------------------------------------------------------------

void ScanOsd::AddLine(const char* text, ...) 
{
  char* buffer = NULL;
  va_list ap;
  va_start(ap, text);
  vasprintf(&buffer, text, ap);
  va_end(ap);

  cOsdItem* item = new cOsdItem(buffer);
  dprint(L_MSG, "%s", buffer);
  free(buffer);

  cOsdMenu::Add(item);
  CursorDown();
  needsUpdate = true;
}


//----------------------------------------------------------------------------

void ScanOsd::UpdateLastLine(const char* text)
{
  char* buffer = NULL;
  asprintf(&buffer, "%s %s", Last()->Text(), text);
  Last()->SetText(buffer, false); // false, so we don't need to free(buffer)
  needsUpdate = true;
  dprint(L_MSG, "%s", text);
}


//----------------------------------------------------------------------------

bool ScanOsd::Parse(const char* line)
{
  char* src = NULL;
  int freq = 0;
  char pol = 0;
  int sr = 0;
      
  bool ok = ( sscanf(line, "%a[^:]:%d :%c :%d ", &src, &freq ,&pol ,&sr) == 4);
  int source = cSource::FromString(src);
  free(src);
      
  pol = toupper(pol);
      
  if (ok && source && freq && sr && (pol == 'H' || pol == 'V' || pol == 'L' || pol == 'R')) 
  {
    if (numSources < MAXSOURCES) {
      sources[numSources]      = source;
      frequency[numSources]    = freq;
      polarization[numSources] = pol;
      symbolRate[numSources]   = sr;
      numSources++;
    }
    return true;
  }

  return false;
}


//----------------------------------------------------------------------------

cChannel* ScanOsd::GetInitial(int i) const
{
  cChannel* c = new cChannel();

  int cr  = MapToDriver(999, CoderateValues);   // Auto
  
#if VDRVERSNUM < 10700
  c->SetSatTransponderData(sources[i], frequency[i], polarization[i], symbolRate[i], cr);
#else
  //int mod = MapToDriver(999, ModulationValues); // Auto 
  int mod = MapToDriver(2, ModulationValues); // QPSK 
  int sys = MapToDriver(0, SystemValues);     // DVB-S
  int ro  = MapToDriver(0, RollOffValues);    // Unknown
  
  c->SetSatTransponderData(sources[i], frequency[i], polarization[i], symbolRate[i], cr, mod, sys, ro);
#endif

  return c; 
}


//----------------------------------------------------------------------------

bool ScanOsd::Tune(cDevice* device, const cChannel* c, int timeout)
{
  const char* mod = NULL;
#if VDRVERSNUM < 10700
  int n = MapToUser(c->Modulation(), ModulationValues);
  switch (n)
  {
    case 0:   mod = "QPSK"; break;
    case 8:   mod = "8PSK"; break;
    case 999: mod = "Auto"; break;
    default:  mod = "----"; break;
  }
 
#else  
  int n = MapToUser(c->Modulation(), ModulationValues, &mod);
  if (n == 999)
    mod = "Auto";
#endif
  
  AddLine("Tuning %d MHz (%s) ...", c->Frequency(), mod);

  if (!device->SwitchChannel(c, false)) {
    UpdateLastLine("Failed");
    return false;  
  }
  
  cCondWait::SleepMs(100);
  
  bool lock = device->HasLock(timeout); 
  UpdateLastLine(lock ? "Success" : "Failed (Timeout)");
  

  return lock;
}


//////////////////////////////////////////////////////////////////////////////


